#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "vector.h"
#include "matrix.h"
#include "pixel.h"
#include "camera.h"
#include "list.h"
#include "material.h"
#include "object.h"
#include "model.h"

// prototypes
void	ray_trace(model_t*,vec_t,vec_t,drgb_t,double);
int	main(int argc, char *argv[]);

int main(int argc, char *argv[])
{
	model_t		*model;
	FILE		*model_file=NULL;

  if(argc != 2) {
    fprintf(stderr,"Usage: %s <model_file>\n",argv[0]);
    exit(1);
  }

  // open model file
  if((model_file = fopen(argv[1],"r")) == NULL) {
     printf("Error opening input file: %s\n",argv[1]);
     return;
  }

  // create model
  model = model_init(model_file);

  // close file
  fclose(model_file);

  // verify that we have the model
  model_print(model,stderr);

	int	i,x,y;
	int	w=model->cam->pixel_dim[0],h=model->cam->pixel_dim[1];
	double	wx,wy,wz=0.0;
	double	ww=model->cam->world_dim[0],wh=model->cam->world_dim[1];
	vec_t	pos,pix,dir;
	drgb_t	color;
	irgb_t	icolor;
	irgb_t	*imgloc;
	irgb_t	*img=NULL;

  // allocate 2D image array and zero out
  assert((img = (irgb_t *)malloc(w*h*sizeof(irgb_t)))!=NULL);
  memset(img,0,w*h*sizeof(irgb_t));

  fprintf(stdout,"P6 %d %d 255\n",w,h);
  for(y=h-1;y>=0;y--) {
    for(x=0;x<w;x++) {
      wx = (double)x/(double)(w-1) * ww;
      wy = (double)y/(double)(h-1) * wh;

      // copy vector v1 to vector v2
      vec_copy(model->cam->view_point,pos);

      // set pixel position vector (in world coordinates)
      pix[0] = wx;  pix[1] = wy;  pix[2] = wz;

      // compute the vector difference  v3 = v2 - v1
      vec_diff(pos,pix,dir);

      // our ray is now {pos, dir} (in world coordinates), normalize dir
      vec_unit(dir,dir);

      // zero out color
      color[0] = 0.0;  color[1] = 0.0;  color[2] = 0.0;

      // trace ray
      ray_trace(model,pos,dir,color,0.0);

      pix_scale(255.0,color,color);

      // where are we in the image (using old i*c + j)
      imgloc = img + y*w + x;
      for(i=0;i<3;i++) (*imgloc)[i] = color[i];
    //fwrite(imgloc,sizeof(irgb_t),1,stdout);

      // instead of using image, just write pixel to stdout
      for(i=0;i<3;i++) icolor[i] = color[i];
      fwrite(icolor,sizeof(irgb_t),1,stdout);
    }
  }

  return(0);
}

void ray_trace(model_t	*model,		// the world (model)
               vec_t	pos,		// the ray pos
               vec_t	dir,		// the ray dir
               drgb_t	color,		// pixel color (output)
               double	raydist)	// distance ray has travelled
{
	double		dis=0.0;
	object_t	*obj=NULL;
	vec_t		hit,N;
	drgb_t		thiscolor = {0.0, 0.0, 0.0};
	drgb_t		ambient,diffuse,specular;

  // get closest object, if any
  if(!(obj = object_find_closest(model->objs,pos,dir,&dis,hit,N)))
    return;

  // if hit, print info
  if(dis > 0) {
#ifdef DEBUG
    fprintf(stderr,"Hit: %-12s ",obj->name);

    fprintf(stderr,"Dist = %8.3lf ",dis);
#endif

    // accumulate distance travelled by ray
    raydist += dis;

    // get object material properties
    material_getamb(obj->mat,ambient);
    material_getdiff(obj->mat,diffuse);
    material_getspec(obj->mat,specular);

    // for now just use ambient color
    pix_copy(ambient,thiscolor);

    // scale by distance
    pix_scale(1.0/raydist,thiscolor,thiscolor);

    // add in this surface color to color
    pix_sum(thiscolor,color,color);
  }
}
