#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "vector.h"
#include "pixel.h"
#include "list.h"
#include "material.h"
#include "object.h"

// create a new object description
void object_init(object_t *obj,FILE *in,list_t *objs,list_t *mats)
{
        // derived class allocates memory, or use object_alloc for generic
        // object_t class---without object_alloc, we are basically simulating           // a pure virtual class (in other words, no one should really
        // be allocating an object_t object)
//      object_t        *obj;

	material_t	*mat;
	char		c,attrname[NAME_LEN],matname[NAME_LEN];
	int		count;

  // create new object and zero it
//assert((obj = (object_t *)malloc(sizeof(object_t))) != NULL);
//memset(obj,0,sizeof(object_t));

  // set cookie
  obj->cookie = OBJ_COOKIE;

  // read descriptive name of object (left_all, center_sphere, etc.)
  assert((count=fscanf(in,"%s",obj->name))==1);

  // consume all chars until we get to '{'
  while((c=fgetc(in)) != EOF && c != '{');

  // set cookie
  obj->cookie = OBJ_COOKIE;

  // read in material (first attribute MUST be material)
  assert((count = fscanf(in,"%s",attrname))==1);

  // read material name (id)
  if(!strcmp(attrname,"material"))
    assert((count = fscanf(in,"%s",matname))==1);

  // find material in material list
  assert((obj->mat=material_getbyname(mats,matname))!=NULL);

  // consume whitespace
  while((c=fgetc(in)) != EOF && c != '\n');

  // initialize default handlers
  obj->printer  = object_print;
  obj->hits     = object_no_hit;
  obj->ambient  = material_getamb;
  obj->diffuse  = material_getdiff;
  obj->specular = material_getspec;

  // add object to object list
  list_add(objs,(void *)obj);
}

// determine if a vector projected from location *base in direction *dir
// hits the object...if so return distance from the base to contact point
double object_no_hit(object_t *obj,vec_t pos,vec_t dir,vec_t hit,vec_t N)
{
  fprintf(stderr,"Object %s failed to provide hit func\n",obj->name);
  return(-1.0);
}

void object_list_print(list_t *objs, FILE *out)
{
	object_t	*obj;

  // reset internal list iterator
  list_reset(objs);

  // print each object; basic list iteration
  while(list_not_end(objs)) {
    obj = (object_t *)list_get_data(objs);
    obj->printer(obj,out);			// invoke polymorphic printer
    list_next_link(objs);
  }
}

void object_print(object_t *obj, FILE *out)
{
  assert(obj->cookie==OBJ_COOKIE);

  // print out 'object' token and object name
  fprintf(out,"%s %s\n",obj->type,obj->name);
  fprintf(out,"{\n");

  // print out material
  fprintf(out,"   %s %s\n","material",material_getname(obj->mat));
}

// name accessor
char* object_getname(object_t *obj)
{
  return obj->name;
}

// find the closest object intersection on the ray with
// at "pos" with direction "dir"
object_t *object_find_closest(list_t *objs,
                              vec_t pos, vec_t dir,
                              double *dis,		// dis to hit point
                              vec_t hit, vec_t N)
{
	// candidate object (as we iterate), distance, hit point, and N
	double		c_dis;
	vec_t		c_hit,c_N;
	object_t	*c_obj=NULL;

	// closest object thus far (and related data, incl. hit, N)
	double		closest_dis=INFINITY;
	vec_t		closest_hit,closest_N;
	object_t	*closest_obj=NULL;

  // check each object; basic list iteration
  for(list_reset(objs); list_not_end(objs); list_next_link(objs)) {

    // get candidate object
    c_obj = (object_t *)list_get_data(objs);

    // if intersection behind us, ignore
    if((c_dis = c_obj->hits(c_obj,pos,dir,c_hit,c_N))<0)
      continue;
    // if distance to hit is smaller, set closest
    else if((0.000001 < c_dis) && (c_dis < closest_dis)) {
      closest_dis = c_dis;
      closest_obj = c_obj;
      vec_copy(c_hit,closest_hit);
      vec_copy(c_N,closest_N);
    } else {
      continue;
    }
  }

  if(closest_obj) {
    vec_copy(closest_hit,hit);
    vec_copy(closest_N,N);
    (*dis) += closest_dis;
  }

  return(closest_obj);
}
